package template

import (
	"bytes"
	"encoding/json"
	"fmt"
	"github.com/andybalholm/cascadia"
	"github.com/charmbracelet/log"
	"gitlab.schukai.com/oss/bob/constants"
	"gitlab.schukai.com/oss/bob/types"
	"gitlab.schukai.com/oss/bob/util"
	"golang.org/x/net/html"
	"golang.org/x/net/html/atom"
	"path"
	"regexp"
	"strings"
)

const attributePrefix = "data-"
const attributeAttributes = "data-attributes"
const attributeReplace = "data-replace"

func removeAttribute(attrs []html.Attribute, key string) []html.Attribute {

	var result []html.Attribute

	for _, attr := range attrs {
		if attr.Key == key {
			continue
		}

		result = append(result, attr)
	}

	return result
}

func setDataAttributesAttribute(node *html.Node, name, attribute, instruction string) {

	value := util.GetAttribute(node.Attr, attributeAttributes)
	if value != "" {
		l := strings.Split(value, ",")
		m := make(map[string]string)
		for _, v := range l {
			v = strings.TrimSpace(v)
			x := strings.Index(v, " ")
			if x > 0 {
				a := v[:x]
				b := v[x+1:]

				if a != attribute {
					m[a] = b
				}
			}
		}
		value = ""
		for k, v := range m {
			if value != "" {
				value += ","
			}
			value += k + " " + v
		}
	}

	node.Attr = removeAttribute(node.Attr, attributeAttributes)

	if value != "" {
		value += ","
	}
	value += attribute + " " + instruction
	node.Attr = append(node.Attr, html.Attribute{Key: name, Val: value})

}

func prepareMeta(node *html.Node, attrKey, attrValue string, storage *types.PageData) {

	if storage.Meta == nil {
		storage.Meta = make(map[string]string)
	}

	if _, ok := storage.Meta[attrValue]; ok {
		return
	}

	sel, err := cascadia.Parse("meta[" + attrKey + "=" + attrValue + "]")
	if err != nil {
		return
	}

	meta := cascadia.Query(node, sel)
	if meta == nil {
		return
	}

	setDataAttributesAttribute(meta, attributeAttributes, "content", "path:meta."+attrValue)

	storage.Meta[attrValue] = util.GetAttribute(meta.Attr, "content")

}

func prepareLanguage(node *html.Node, storage *types.PageData) {

	selector, err := cascadia.Parse("html")
	if err != nil {
		return
	}

	n := cascadia.Query(node, selector)
	if n == nil {
		return
	}

	setDataAttributesAttribute(n, attributeAttributes, "lang", "path:lang")

	lang := util.GetAttribute(n.Attr, "lang")
	if lang == "" {
		lang = "en"
	}

	storage.Lang = lang

}

func prepareTitle(node *html.Node, storage *types.PageData) {

	if storage.Title != "" {
		return
	}

	selector, err := cascadia.Parse("title")
	if err != nil {
		return
	}

	n := cascadia.Query(node, selector)
	if n == nil {
		return
	}

	n.Attr = removeAttribute(node.Attr, attributeReplace)
	n.Attr = append(node.Attr, html.Attribute{Key: attributeReplace, Val: "path:title"})

	title := ""
	for c := n.FirstChild; c != nil; c = c.NextSibling {
		if c.Type == html.TextNode {
			title += c.Data
		}
	}

	storage.Title = title

}

func prepareAnchors(node *html.Node, storage *types.PageData) {
	selector, err := cascadia.Parse("a")
	if err != nil {
		return
	}

	list := cascadia.QueryAll(node, selector)
	if list == nil {
		return
	}

	for _, n := range list {

		logContent := buildLogContent(n, "")
		log.Info("Found a node: " + logContent)

		title := util.GetAttribute(n.Attr, "title")
		hreflang := util.GetAttribute(n.Attr, "hreflang")
		href := util.GetAttribute(n.Attr, "href")

		k := title + hreflang + href
		if k == "" {
			log.Warn("The anchor '" + logContent + "' has no title, hreflang and href attribute! Will use the content of the anchor as a key.")
			k = logContent
		}

		id := util.GetOrCreateReference(n, k)
		if id == "" {
			id = util.GetNextId()
		}

		// check if id is already in use
		found := false
		sameContent := false
		for _, t := range storage.Anchors {
			if t.Id == id {
				// check if the content is the same
				if t.Title == title && t.HrefLang == hreflang && t.Href == href {
					sameContent = true
					continue
				}
				found = true
				break
			}
		}

		if sameContent {
			log.Info("The ID " + id + " for the anchor '" + logContent + "' is already in use and the content is the same! Will use this ID for the anchor.")
		} else if found {
			originalId := id
			id = util.GetNextId()
			log.Warn("We want to use the ID " + originalId + " for the anchor '" + logContent + "', but this ID is already in use! " +
				"This could be a problem if you want to use the anchor in a translation. " +
				"Please change the anchor or add the attribute '" + constants.DataBobReferenceAttributeKey + "' to the tag. " +
				"Will use now use ID: " + id + " instead.")
		}

		setDataAttributesAttribute(n, attributeAttributes, "href", "path:anchors."+id+".href")
		setDataAttributesAttribute(n, attributeAttributes, "title", "path:anchors."+id+".title")
		setDataAttributesAttribute(n, attributeAttributes, "hreflang", "path:anchors."+id+".hreflang")

		n.Attr = removeAttribute(n.Attr, constants.DataBobReferenceAttributeKey)
		n.Attr = append(n.Attr, html.Attribute{Key: constants.DataBobReferenceAttributeKey, Val: id})

		if sameContent {
			continue
		}

		log.Info("Added anchor: " + id + " to storage")

		storage.Anchors = append(storage.Anchors, types.Anchor{
			Id:       id,
			Title:    title,
			HrefLang: hreflang,
			Href:     href,
		})

	}

}

func prepareImages(node *html.Node, storage *types.PageData) {

	selector, err := cascadia.Parse("img")
	if err != nil {
		return
	}

	list := cascadia.QueryAll(node, selector)
	if list == nil {
		return
	}

	for _, n := range list {

		logContent := buildLogContent(n, "")
		log.Info("Found img node: " + logContent)

		alt := util.GetAttribute(n.Attr, "alt")
		title := util.GetAttribute(n.Attr, "title")
		source := util.GetAttribute(n.Attr, "src")

		id := util.GetOrCreateReference(n, alt+title+source)
		if id == "" {
			id = util.GetNextId()
		}

		// check if id is already in use
		found := false
		sameContent := false
		for _, t := range storage.Images {
			if t.Id == id {
				// check if the content is the same
				if t.Alt == alt && t.Title == title && t.Source == source {
					sameContent = true
					continue
				}
				found = true
				break
			}
		}

		if sameContent {
			log.Info("The ID " + id + " for the image '" + logContent + "' is already in use and the content is the same! Will use this ID for the image.")
		} else if found {
			originalId := id
			id = util.GetNextId()
			log.Warn("We want to use the ID " + originalId + " for the image '" + logContent + "', but this ID is already in use! " +
				"This could be a problem if you want to use the image in a translation. " +
				"Please change the image or add the attribute '" + constants.DataBobReferenceAttributeKey + "' to the tag. " +
				"Will use now use ID: " + id + " instead.")
		}

		n.Attr = removeAttribute(n.Attr, constants.DataBobReferenceAttributeKey)
		n.Attr = append(n.Attr, html.Attribute{Key: constants.DataBobReferenceAttributeKey, Val: id})

		setDataAttributesAttribute(n, attributeAttributes, "src", "path:content."+id+".src")
		setDataAttributesAttribute(n, attributeAttributes, "alt", "path:content."+id+".alt")
		setDataAttributesAttribute(n, attributeAttributes, "title", "path:content."+id+".title")

		if sameContent {
			continue
		}

		storage.Images = append(storage.Images, types.Image{
			Id:     id,
			Alt:    alt,
			Title:  title,
			Source: source,
		})

	}

}

func prepareTranslationJson(node *html.Node, storage *types.PageData) {
	selector, err := cascadia.Parse("script[data-monster-role=translations]")
	if err != nil {
		return
	}

	list := cascadia.QueryAll(node, selector)
	if list == nil {
		return
	}

	for _, n := range list {

		id := util.GetAttribute(n.Attr, constants.DataBobReferenceAttributeKey)
		if id == "" {
			id = util.GetNextId()
		}

		log.Info("Found json translation: " + id)

		found := false
		for _, t := range storage.Translations {
			if t.Id == id {
				found = true
				break
			}
		}

		if found {
			log.Info("The ID " + id + " for the json translation is already in use! If this is not intended, please change the ID in the HTML file. Will use ID: " + id)
			continue
		}

		typ_ := util.GetAttribute(n.Attr, "type")

		n.Attr = removeAttribute(n.Attr, attributeReplace)
		n.Attr = append(n.Attr, html.Attribute{Key: attributeReplace, Val: "path:translations." + id + ".content"})

		content := ""
		for c := n.FirstChild; c != nil; c = c.NextSibling {
			if c.Type == html.TextNode {
				content += c.Data
			}
		}

		content = strings.TrimSpace(content)

		t := make(map[string]any)
		err := json.Unmarshal([]byte(content), &t)
		if err != nil {
			fmt.Println(err)
		}

		log.Info("Added json translation: " + id + " (" + typ_ + ") to storage")
		storage.Translations = append(storage.Translations, types.Translations{
			Id:        id,
			Type:      typ_,
			KeyValues: t,
		})

	}

}

func prepareTextNodes(node *html.Node, storage *types.PageData) {

	selector, err := cascadia.Parse("body")
	if err != nil {
		return
	}

	body := cascadia.Query(node, selector)
	if body == nil {
		return
	}

	runNodes(body, storage)

}

func runNodes(n *html.Node, storage *types.PageData) {

	nodeList := []*html.Node{}
	for c := n.FirstChild; c != nil; c = c.NextSibling {
		nodeList = append(nodeList, c)
	}

	for _, n := range nodeList {
		checkNodes(n, storage)
		runNodes(n, storage)
	}

}

func renderNode(n *html.Node) string {
	var buf bytes.Buffer
	html.Render(&buf, n)
	return buf.String()
}

func removeDuplicateWhitespaces(str string) string {
	re := regexp.MustCompile(`\s+`)
	return re.ReplaceAllString(str, " ")
}

func handleTextNode(n *html.Node, storage *types.PageData) {
	content := strings.TrimSpace(n.Data)

	if content == "" {
		return
	}

	logContent := buildLogContent(n.Parent, content)
	log.Info("Found text node: " + logContent)

	if n.Parent != nil {
		if n.Parent.Type == html.ElementNode {

			// if span and has constants.DataBobReferenceAttributeKey, ignore
			if n.Parent.Data == "span" && util.HasAttribute(n.Parent.Attr, constants.DataBobReferenceAttributeKey) {

				id := util.GetAttribute(n.Parent.Attr, constants.DataBobReferenceAttributeKey)

				log.Info("Added text: " + id + " (" + logContent + ") to storage")
				storage.Text = append(storage.Text, types.Text{
					Id:   id,
					Text: content,
				})

				return
			}

			if util.HasAttribute(n.Parent.Attr, "data-monster-head") {
				return
			}
		}
	}

	id, err := util.BuildTextKey(content)
	if err != nil || id == "" {
		id = util.GetNextId()
		log.Warn("Could not build a text key for: '" + logContent + "', using ID: " + id + " instead.")
	} else {
		log.Info("Create a text key: " + id)
	}

	// check if id is already in use
	found := false
	sameContent := false
	for _, t := range storage.Text {
		if t.Id == id {
			// check if the content is the same
			if t.Text == content {
				sameContent = true
				continue
			}
			found = true
			break
		}
	}

	if sameContent {
		log.Info("The ID " + id + " for the text '" + logContent + "' is already in use and the content is the same! Will use this ID for the text.")
	} else if found {
		originalId := id
		id = util.GetNextId()
		log.Warn("We want to use the ID " + originalId + " for the text '" + logContent +
			"', but this ID is already in use! " +
			"This could be a problem if you want to use the text in a translation. " +
			"Please change the text or the ID in the HTML file. " +
			"Will use now use ID: " + id + " instead.")
	}

	parent := n.Parent
	span := &html.Node{
		Type: html.ElementNode,
		Data: atom.Span.String(),
		Attr: []html.Attribute{
			{
				Key: constants.DataBobReferenceAttributeKey,
				Val: id,
			},
			{
				Key: "data-replace-self",
				Val: "path:text." + id + ".text",
			},
		},
	}

	parent.InsertBefore(span, n)
	parent.RemoveChild(n)

	span.AppendChild(n)

	if sameContent {
		return
	}

	log.Info("Added text: " + id + " (" + logContent + ") to storage")
	storage.Text = append(storage.Text, types.Text{
		Id:   id,
		Text: content,
	})
}

func buildLogContent(n *html.Node, content string) string {
	logContent := ""
	if len(content) == 0 {
		logContent = "Struct: " + renderNode(n)
	} else if len(content) < 10 {
		logContent = "Parent struct for (" + content + "): " + renderNode(n)

	} else {
		logContent = content
	}

	if len(logContent) > 100 {
		logContent = logContent[:100] + "..."
	}

	logContent = removeDuplicateWhitespaces(logContent)
	return logContent
}

func checkNodes(n *html.Node, storage *types.PageData) {
	if n.Parent != nil {
		if n.Parent.Type == html.ElementNode {
			if n.Parent.Data == "script" || n.Parent.Data == "style" {
				return
			}
		}
	}

	if n.Type == html.TextNode {
		handleTextNode(n, storage)
	} else if n.Type == html.ElementNode {
		switch n.Data {
		case "monster-datatable":
			checkMonsterDatatableHead(n, storage)
		}

	}

}

func checkMonsterDatatableHead(n *html.Node, storage *types.PageData) {
	selector, err := cascadia.Parse("[data-monster-head]")
	if err != nil {
		return
	}

	list := cascadia.QueryAll(n, selector)
	if list == nil {
		return
	}

	for _, div := range list {

		for i, attr := range div.Attr {
			if attr.Key == constants.DataBobReferenceAttributeKey && attr.Val == "" {
				div.Attr = append(div.Attr[:i], div.Attr[i+1:]...)
			}
		}

		head := util.GetAttribute(div.Attr, "data-monster-head")

		trimHead := strings.TrimSpace(head)

		if trimHead == "" {
			log.Warn("The head '" + buildLogContent(div, "") + "' has no content! Will use the content of the head as a key.")
			continue
		}

		logContent := buildLogContent(div, head)

		id := util.GetOrCreateReference(div, head)
		if id == "" {
			id = util.GetNextId()
			log.Warn("Could not build key for: '" + logContent + "', using ID: " + id + " instead.")
		} else {
			log.Info("Create key: " + id)
		}

		// check if id is already in use
		found := false
		sameContent := false
		for _, t := range storage.Text {
			if t.Id == id {
				// check if the content is the same
				if t.Text == head {
					sameContent = true
					continue
				}
				found = true
				break
			}
		}

		if sameContent {
			log.Info("The ID " + id + " for the head '" + logContent + "' is already in use and the content is the same! Will use this ID for the head.")
		} else if found {
			originalId := id
			id = util.GetNextId()
			log.Warn("We want to use the ID " + originalId + " for the head '" + logContent +
				"', but this ID is already in use! " +
				"This could be a problem if you want to use the head in a translation. " +
				"Please change the head or add the attribute '" + constants.DataBobReferenceAttributeKey + "' to the tag. " +
				"Will use now use ID: " + id + " instead.")

		}

		div.Attr = removeAttribute(div.Attr, "data-attributes")
		div.Attr = append(div.Attr, html.Attribute{Key: "data-attributes", Val: "data-monster-head path:text." + id + ".text"})

		if sameContent {
			continue
		}

		log.Info("Added head: " + id + " (" + logContent + ") to storage")
		storage.Text = append(storage.Text, types.Text{
			Id:   id,
			Text: head,
		})
	}
}

func PrepareHtmlFile(from, to string, storage types.PageDataStorage) (string, error) {
	node, err := util.LoadHtml(from)
	if err != nil {
		return "", err
	}

	var pd *types.PageData
	var ok bool

	p := path.Base(from)

	if pd, ok = storage[p]; !ok {
		pd = types.NewPageData()
		storage[p] = pd
	}

	prepareLanguage(node, pd)
	prepareTitle(node, pd)
	prepareMeta(node, "name", "description", pd)
	prepareMeta(node, "name", "keywords", pd)
	prepareMeta(node, "name", "author", pd)

	prepareImages(node, pd)
	prepareAnchors(node, pd)
	prepareTextNodes(node, pd)

	prepareTranslationJson(node, pd)

	pd.Export = path.Join(pd.Lang, path.Base(from))
	to = path.Join(to, path.Base(from))

	return p, util.SaveHtml(to, node)

}
