package translate

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"os"
	"strings"
)

// translateTextsOpenAI übersetzt eine Liste von Texten mittels der OpenAI API.
// Es wird für jeden Text ein eigener API-Aufruf gemacht.
// Beachte: Für jeden API-Aufruf muss der OPENAI_API_KEY in der Umgebung gesetzt sein.
func translateTextsOpenAI(texts []string, targetLang string) (map[string]string, error) {
	apiKey := os.Getenv("OPENAI_API_KEY")
	if apiKey == "" {
		return nil, errors.New("Missing OPENAI_API_KEY environment variable")
	}

	translations := make(map[string]string)

	for _, text := range texts {
		// Leere Texte überspringen.
		if strings.TrimSpace(text) == "" {
			translations[text] = text
			continue
		}

		// Prompt formulieren: Übersetze den Text von Englisch in die Zielsprache.
		prompt := fmt.Sprintf("Translate the following text from English to %s:\n\n%s", targetLang, text)

		// Anfrage-Payload aufbauen.
		requestBody := map[string]interface{}{
			"model":       "text-davinci-003",
			"prompt":      prompt,
			"max_tokens":  1024,
			"temperature": 0.3,
		}

		requestBytes, err := json.Marshal(requestBody)
		if err != nil {
			return nil, err
		}

		// HTTP-Request an die OpenAI API.
		req, err := http.NewRequest("POST", "https://api.openai.com/v1/completions", bytes.NewBuffer(requestBytes))
		if err != nil {
			return nil, err
		}
		req.Header.Set("Content-Type", "application/json")
		req.Header.Set("Authorization", "Bearer "+apiKey)

		client := &http.Client{}
		resp, err := client.Do(req)
		if err != nil {
			return nil, err
		}
		defer resp.Body.Close()

		if resp.StatusCode != http.StatusOK {
			body, _ := io.ReadAll(resp.Body)
			return nil, fmt.Errorf("OpenAI API returned status %d: %s", resp.StatusCode, body)
		}

		// Antwort-JSON dekodieren.
		var responseData struct {
			Choices []struct {
				Text string `json:"text"`
			} `json:"choices"`
		}
		if err := json.NewDecoder(resp.Body).Decode(&responseData); err != nil {
			return nil, err
		}

		if len(responseData.Choices) == 0 {
			return nil, fmt.Errorf("no translation received for text: %s", text)
		}

		// Übersetzung aus der Antwort extrahieren.
		translation := strings.TrimSpace(responseData.Choices[0].Text)
		translations[text] = translation
	}

	return translations, nil
}
